/*
   pam_bioapi - Provides a PAM-compliant interface to use in biometrically authenticating local users.

   Copyright (C) 2005 Michael R. Crusoe <michael@qrivy.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define _GNU_SOURCE
#define PAM_SM_AUTH /* we provide authentication functions */
#define PAM_SM_PASSWORD /* we provide "password" setting functions, in our case enrollment of the biometric */
#include <security/_pam_macros.h>
#include <security/pam_modules.h>
#include <bioapi.h>
#include <bioapi_util.h>
#include <bioapi_uuid.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/stat.h>

#define DEFBIRDBPATH "/etc/bioapi/pam/"

int readBIR(const char *BIRUUID, const char *birDbPath, const char *user, BioAPI_INPUT_BIR *BIR);
int writeBIR(BioAPI_BIR_HANDLE bir, BioAPI_HANDLE bspHandle, const char *birDbPath, const char *biruuid, const char *user);

static const struct bioapi_version version = { 1, 10 };

/* Authentication functions */

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags,
                                   int argc, const char **argv) {
	int pamRet;
	BioAPI_RETURN bRet;
	const char *user;
	const char *uuidString;
	const char *birDbPath;
	const BioAPI_FAR maxFAR = 1;
	BioAPI_HANDLE bspHandle;
	BioAPI_INPUT_BIR BIR;
	BioAPI_BOOL result = BioAPI_FALSE;
	BioAPI_FAR FARAchieved;
	BioAPI_UUID tempUuid;
	const BioAPI_UUID *uuid;
	int tries;
	
	openlog("pam_bioapi", LOG_PID | LOG_CONS, LOG_AUTHPRIV);
	
	pamRet = pam_fail_delay(pamh, 100);
	
	pamRet = pam_get_user(pamh, &user, NULL);
	if (pamRet != PAM_SUCCESS || strlen(user) == 0) {
		if (pamRet == PAM_CONV_AGAIN) {
			pamRet = PAM_INCOMPLETE;
		} else {
			syslog(LOG_ALERT, "Error retrieving user name: %s", pam_strerror(pamh, pamRet));
			pamRet = PAM_USER_UNKNOWN;
		}
		closelog();
		return pamRet;
	}
	
	if (argc > 0) {
		uuidString = argv[0];
		if (argc > 1) {
			birDbPath = argv[1];
		} else {
			birDbPath = DEFBIRDBPATH;
		}
	} else {
		syslog(LOG_ALERT, "Mandatory BSP UUID not provided");
		closelog();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bRet = BioAPI_Init(&version, 0, NULL, 0, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to initilize BioAPI framework, BioAPI error #:%x.", bRet);
		closelog();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bRet = BioAPI_GetStructuredUUID(uuidString, &tempUuid);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to parse UUID (BioAPI error #:%x) : %s", bRet, uuidString);
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	uuid = (const BioAPI_UUID *) malloc(sizeof(BioAPI_UUID));
	if (uuid == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory.");
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	memcpy(uuid, tempUuid, sizeof(BioAPI_UUID));
	bRet = BioAPI_ModuleLoad(uuid, 0, NULL, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to load BioAPI BSP with UUID of %s, BioAPI error #%x.", uuidString, bRet);
		closelog();
		free(uuid);
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bRet = BioAPI_ModuleAttach(uuid, &version, &BioAPIMemoryFuncs, 0, 0, 0, 0, NULL, 0, NULL, &bspHandle);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to attach default device to BioAPI BSP with UUID of %s, BioAPI error #%x.", uuidString, bRet);
		closelog();
		(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
		free(uuid);
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	pamRet = readBIR(uuidString, birDbPath, user, &BIR);
	if (pamRet != 0) {
		closelog();
		(void)BioAPI_ModuleDetach(bspHandle);
		(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
		free(uuid);
		(void)BioAPI_Terminate();
		return pamRet;
	}

	tries = 3;
	do {
		bRet = BioAPI_Verify(bspHandle, &maxFAR, NULL, NULL, &BIR, NULL, &result, &FARAchieved, NULL, NULL, -1, NULL);
		switch(bRet) {
			case BioAPI_OK:
				if (result == BioAPI_TRUE) {
					tries = 0;
				} else {
					tries--;
				}
				break;
			case BioAPIERR_BSP_UNABLE_TO_CAPTURE:
				tries--;
				break;
			case BioAPIERR_BSP_BIR_SIGNATURE_FAILURE:
				syslog(LOG_ALERT, "Error in BIR signature verification while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
			case BioAPIERR_BSP_TOO_MANY_HANDLES:
				syslog(LOG_ALERT, "Ran out of handles while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
			case BioAPIERR_BSP_TIMEOUT_EXPIRED:
				tries--;
				break;
			case BioAPIERR_BSP_INCONSISTENT_PURPOSE:
				syslog(LOG_ALERT, "BSP Error: Inconsistent Purpose -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
			case BioAPIERR_BSP_RECORD_NOT_FOUND:
				syslog(LOG_ALERT, "BSP Error: Record not found -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
			default:
				syslog(LOG_ALERT, "Unanticipated BioAPI error#:%x", bRet);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
		}
	} while (tries > 0);
	
	(void)BioAPI_ModuleDetach(bspHandle);
	(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
	free(uuid);
	(void)BioAPI_Terminate();
	closelog();
	if (result == BioAPI_TRUE) {
		return PAM_SUCCESS;
	} else {
		return PAM_AUTH_ERR;
	}
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags,
			      int argc, const char **argv) {
	return PAM_SUCCESS;
}

/* "Password" functions (in our case, enrollment of the biometric) */

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv) {
	int pamRet;
	BioAPI_RETURN bRet;
	const char *user;	
	const char *uuidString;
	const char *birDbPath;
	BioAPI_UUID tempUuid;
	const BioAPI_UUID *uuid;
	BioAPI_HANDLE bspHandle;
	BioAPI_BIR_HANDLE bir;
	int tries;

	openlog("pam_bioapi", LOG_PID | LOG_CONS, LOG_AUTHPRIV);
	
	pamRet = pam_fail_delay(pamh, 100);
	if ((flags & PAM_PRELIM_CHECK) == PAM_PRELIM_CHECK) {
		closelog();
		return PAM_SUCCESS; /* No remote resource to check */
	}
	
	pamRet = pam_get_user(pamh, &user, NULL);
	if (pamRet != PAM_SUCCESS || strlen(user) == 0) {
		if (pamRet == PAM_CONV_AGAIN) {
			pamRet = PAM_INCOMPLETE;
		} else {
			syslog(LOG_ALERT, "Error retrieving user name: %s", pam_strerror(pamh, pamRet));
			pamRet = PAM_USER_UNKNOWN;
		}
		closelog();
		return pamRet;
	}
	
	if (argc > 0) {
		uuidString = argv[0];
		if (argc > 1) {
			birDbPath = argv[1];
		} else {
			birDbPath = DEFBIRDBPATH;
		}
	} else { 
		syslog(LOG_ALERT, "Mandatory BSP UUID not provided");
		closelog();
		return PAM_AUTHTOK_ERR;
	}
	
	bRet = BioAPI_Init(&version, 0, NULL, 0, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to initilize BioAPI framework, BioAPI error #:%x.", bRet);
		closelog();
		return PAM_AUTHTOK_ERR;
	}
	
	bRet = BioAPI_GetStructuredUUID(uuidString, &tempUuid);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to parse UUID (BioAPI error #:%x) : %s", bRet, uuidString);
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}

	uuid = (const BioAPI_UUID *) malloc(sizeof(BioAPI_UUID));
	if (uuid == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory.");
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}
	
	memcpy(uuid, tempUuid, sizeof(BioAPI_UUID));
	bRet = BioAPI_ModuleLoad(uuid, 0, NULL, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to load BioAPI BSP with UUID of %s, BioAPI error #%x.", uuidString, bRet);
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}

	bRet = BioAPI_ModuleAttach(uuid, &version, &BioAPIMemoryFuncs, 0, 0, 0, 0, NULL, 0, NULL, &bspHandle);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to attach default device to BioAPI BSP with UUID of %s, BioAPI error #%x.", uuidString, bRet);
		closelog();
		(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
		free(uuid);
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}
	
	tries = 3;
	do {
		bRet = BioAPI_Enroll(bspHandle, BioAPI_PURPOSE_ENROLL_FOR_VERIFICATION_ONLY, NULL, &bir, NULL, -1, NULL);
		switch (bRet) {
			case BioAPI_OK:
				tries = 0;
				pamRet = writeBIR(bir, bspHandle, birDbPath, uuidString, user);
				break;
			case BioAPIERR_BSP_UNABLE_TO_CAPTURE:
			case BioAPIERR_BSP_TIMEOUT_EXPIRED:
				tries--;
				break;
			case BioAPIERR_BSP_TOO_MANY_HANDLES:
			case BioAPIERR_BSP_PURPOSE_NOT_SUPPORTED:
			case BioAPIERR_BSP_RECORD_NOT_FOUND:
			default:
				syslog(LOG_ALERT, "Unable to enroll user %s using BSP with UUID of %s, BioAPI error #%x.", user, uuidString, bRet);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHTOK_ERR;
		}
	} while (tries > 0);
	(void)BioAPI_ModuleDetach(bspHandle);
	(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
	free(uuid);
	(void)BioAPI_Terminate();
	closelog();
	if (bRet != PAM_SUCCESS) {
		return PAM_AUTHTOK_ERR;
	} else {
		return bRet;
	}
}

int readBIR(const char *biruuid, const char *birDbPath, const char *user, BioAPI_INPUT_BIR *bir) {
	FILE *inputFile;
	char filename[strlen(birDbPath) + strlen(biruuid) + 1 + strlen(user) + 4 + 1];
	size_t bytes, bytesRead;
	BioAPI_BIR *tempBir;

	strcpy(filename, birDbPath);
	strcat(filename, biruuid);
	strcat(filename, "/");
	strcat(filename, user);
	strcat(filename, ".bir");
	
	inputFile = fopen (filename, "r");
	if (inputFile == NULL) {
		syslog(LOG_ALERT, "Unable to open biometric identification record, \"%s\", for user \"%s\"", filename, user);
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bytes = sizeof(BioAPI_BIR_HEADER);
	tempBir = (BioAPI_BIR *)malloc(sizeof(BioAPI_BIR));
	if (tempBir == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory.");
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bytesRead = fread(&(tempBir->Header), bytes, 1, inputFile);
	bytes = (tempBir->Header.Length) - sizeof(BioAPI_DATA);
	tempBir->BiometricData = (BioAPI_BIR_BIOMETRIC_DATA_PTR)malloc(bytes);
	if (tempBir->BiometricData == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory.");
		free(tempBir);
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bytesRead = fread(tempBir->BiometricData, bytes, 1, inputFile);
	bytes = 0;
	bytesRead = fread(&bytes, 4, 1, inputFile);
	
	if (&bytesRead != 0 && &bytes != 0) {
		tempBir->Signature = (BioAPI_DATA_PTR)malloc(sizeof(BioAPI_DATA));
		if (tempBir->Signature == 0) {
			syslog(LOG_ALERT, "Unable to allocate memory.");
			free(tempBir->BiometricData);
			free(tempBir);
			return PAM_AUTHINFO_UNAVAIL;
		}	
		tempBir->Signature->Length = (uint32)bytes;
		tempBir->Signature->Data = (uint8 *)malloc(bytes);
		if (tempBir->Signature->Data == 0) {
			syslog(LOG_ALERT, "Unable to allocate memory.");
			free(tempBir->BiometricData);
			free(tempBir->Signature);
			free(tempBir);
			return PAM_AUTHINFO_UNAVAIL;
		}	
		bytesRead = fread(tempBir->Signature->Data, bytes, 1, inputFile);
	}
	(void)fclose(inputFile);
	bir->InputBIR.BIR = tempBir;
	bir->Form = BioAPI_FULLBIR_INPUT;
	return PAM_SUCCESS;
}

int writeBIR(BioAPI_BIR_HANDLE bir, BioAPI_HANDLE bspHandle, const char *birDbPath, const char *biruuid, const char *user) {
	FILE *outputFile;
	char filename[strlen(birDbPath) + strlen(biruuid) + 1 + strlen(user) + 4 + 1];
	struct stat status;
	BioAPI_BIR_PTR birData;
	BioAPI_RETURN bRet;

	strcpy(filename, birDbPath);
	(void)stat(filename, &status); /* Does the birDb directory exist? */
	if (errno == ENOENT) { /* No. */
		if (mkdir(filename, S_IRWXU | S_IRWXG) == -1) { /* Try to make it */
			return PAM_AUTHTOK_ERR; /* But fail out if we can't. */
		}
	}
	
	strcat(filename, biruuid);
	(void)stat(filename, &status); /* Does the BSP specific directory exist? */
	if (errno == ENOENT) { /* No. */
		if (mkdir(filename, S_IRWXU | S_IRWXG) == -1) { /* Try to make it */
			return PAM_AUTHTOK_ERR; /* But fail out if we can't */
		}
	}
	
	strcat(filename, "/");
	strcat(filename, user);
	strcat(filename, ".bir");
	
	birData = NULL;
	bRet = BioAPI_GetBIRFromHandle(bspHandle, bir, &birData);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to write biometric identification record, \"%s\": BioAPI error #%x", filename, bRet);
		return PAM_AUTHTOK_ERR;
	}
	
	outputFile = fopen(filename, "w+");
	if (outputFile == NULL) {
		return PAM_AUTHTOK_ERR;
	}
	
	(void)fwrite(&(birData->Header), sizeof(BioAPI_BIR_HEADER), 1, outputFile);
	(void)fwrite(birData->BiometricData, birData->Header.Length - sizeof(BioAPI_BIR_HEADER), 1, outputFile);

	if (birData->Signature) {
		(void)fwrite(&(birData->Signature->Length), 4, 1, outputFile);
		(void)fwrite(birData->Signature->Data, (size_t)birData->Signature->Length, 1, outputFile);
	}

	(void)fclose(outputFile);
	free(birData->BiometricData);

	if (birData->Signature) {
		free(birData->Signature->Data);
		free(birData->Signature);
	}
	
	free(birData);

	return PAM_SUCCESS;
}
